#pragma once

#include <boost/asio/io_service.hpp>
#include <boost/asio/read.hpp>
#include <boost/chrono.hpp>
#include "gtest/gtest.h"

namespace ds {
namespace gtest {
namespace asio  {

/**
 * Duration for timers that are not expected to fire during test.
 * The whole test is expected to finish before this duration expires.
 */
static constexpr boost::chrono::milliseconds Timeout = boost::chrono::seconds(10);

/**
 * Duration for timers that are expected to fire during test.
 */
static constexpr boost::chrono::milliseconds Latency = boost::chrono::milliseconds(12);

/** Run io_service up to duration period until it is stopped.
 * @return true when duration period expired.
 */
::testing::AssertionResult runExpireFromNow(
    boost::asio::io_service& ioService,
    boost::chrono::milliseconds duration
);

/** Run io_service and expect stop within duration. Duration defaults to Timeout / 2. */
::testing::AssertionResult runStop(boost::asio::io_service &ioService, boost::chrono::milliseconds duration = Timeout / 2);
/** Handy macro */
#define ASSERT_RUN_STOP(ioService) \
    ASSERT_TRUE(::ds::gtest::asio::runStop((ioService)))


/** Run io_service and expect no stop within duration. Duration defaults to Latency / 2. */
::testing::AssertionResult runNoStop(boost::asio::io_service &ioService, boost::chrono::milliseconds duration = Latency / 2);
/** Handy macro */
#define ASSERT_RUN_NO_STOP(ioService) \
    ASSERT_TRUE(::ds::gtest::asio::runNoStop((ioService)))

/**
 * Run io_service and do not expect stop within first Latency / 2 duration
 * and do expect stop within next Latency duration.
 *
 * Checks that some timer does not expire immediately but after Latency period.
 */
::testing::AssertionResult runLatencyStop(boost::asio::io_service &ioService);
#define ASSERT_RUN_LATENCY_STOP(ioService) \
    ASSERT_TRUE(::ds::gtest::asio::runLatencyStop((ioService)))

/**
 * Like asio::read, but times out using \ref runStop().
 */
template<typename SyncReadStream, typename MutableBufferSequence>
std::size_t runRead(SyncReadStream& s, const MutableBufferSequence& buffers)
{
    boost::system::error_code e;
    size_t n = 0;
    boost::asio::async_read(s, buffers, [&](const boost::system::error_code& ee, std::size_t nn)
    {
        if (e == boost::asio::error::operation_aborted) {
            return;
        }
        n = nn;
        e = ee;
        s.get_io_service().stop();
    });

    if (!runStop(s.get_io_service())) {
        throw std::runtime_error("Timeout");
    }
    if (e) {
        throw boost::system::system_error(e);
    }
    return n;
}

}}} //namespace ds::gtest::asio
